# pylint: disable=too-few-public-methods
import inspect
import json
import logging
import re

import flask
import flask.views
import werkzeug.urls
from werkzeug.wrappers import Response as ResponseBase

import sepiida.errors
import sepiida.parsing
import sepiida.serialization
from sepiida import fields
from sepiida.errors import APIException, Error

LOGGER = logging.getLogger(__name__)

class APIEndpoint(flask.views.View):
    ENDPOINT       = None
    ERRORS         = []
    LIST_ENVELOPE  = 'resources'
    PUBLIC_METHODS = []
    SIGNATURE      = None
    CACHING        = {'GET': 'max-age=10'}

    def __init__(self):
        self.fields = []
        self.filters = {}

    @classmethod
    def endpoint_with_id(cls):
        return cls.ENDPOINT + '<int:_id>/'

    @classmethod
    def endpoint_with_uuid(cls):
        return cls.ENDPOINT + '<uuid:uuid>/'

    def _map_request_to_method(self, request, single_resource):
        if single_resource:
            if request.method == 'POST':
                raise AttributeError("POSTs not allowed on single resource URLs")
            return getattr(self, request.method.lower())
        else:
            if request.method == 'GET':
                return getattr(self, 'list')
            elif request.method == 'PUT':
                raise AttributeError("PUTs are not allowed on collection URLs")
            else:
                return getattr(self, request.method.lower())

    def _fixup_response(self, request, response):
        if isinstance(response, ResponseBase):
            return response, response.status_code, response.headers
        elif isinstance(response, tuple):
            return response
        elif response is None:
            headers = {
                'Content-Type': 'text/plain',
                'Cache-Control': self.CACHING.get(request.method)
            }
            return flask.make_response('', 204, headers), 204, headers
        elif request.method == 'POST':
            return response, 201, {}
        else:
            return response, 200, {}

    def _convert_exception(self, e):
        for handler_spec in self.get_error_handlers():
            try:
                raise handler_spec.translate(e)
            except TypeError:
                continue

    @staticmethod
    def _filter_fields(selected_fields, response):
        if not selected_fields:
            return response
        return {k: response[k] for k in selected_fields.values}

    def _apply_envelope(self, response):
        envelope_signature = fields.Object(s={
            self.LIST_ENVELOPE: fields.Array(s=self.SIGNATURE)
        })
        return envelope_signature.package({self.LIST_ENVELOPE: response})

    def _package_response(self, request, method, response, selected_fields):
        response, status, headers = self._fixup_response(flask.request, response)
        if 'Cache-Control' not in headers:
            headers['Cache-Control'] = self.CACHING.get(request.method)
        if response is None or response == '':
            response, headers = sepiida.serialization.prepare_response(flask.request, response, headers, None)
            return flask.make_response('', 204, headers)

        if not isinstance(response, ResponseBase):
            if self.SIGNATURE:
                if method == getattr(self, 'list', None):
                    response = self._apply_envelope(response)
                else:
                    response = self.SIGNATURE.package(response)
                response = self._filter_fields(selected_fields, response)
            response, headers = sepiida.serialization.prepare_response(flask.request, response, headers, self.SIGNATURE)
            response = flask.make_response(response, status, headers)
        return response

    def get_error_handlers(self):
        handlers = [getattr(parent, 'ERRORS', []) for parent in self.__class__.__mro__]
        handlers = getattr(self, 'ERRORS', []) + [handler for parent_handlers in handlers for handler in parent_handlers]
        return [handler for handler in handlers if isinstance(handler, sepiida.errors.Specification)]

    @classmethod
    def authenticate(cls):
        return

    @staticmethod
    def get_arguments(request):
        if request.method == 'GET':
            body = request.data
            if not body:
                return request.args
            try:
                return werkzeug.urls.url_decode(body, charset='utf-8', errors='strict')
            except UnicodeDecodeError as e:
                raise sepiida.parsing.ParseErrors(errors=[sepiida.parsing.ParseError(error_code='bad-querystring-body', message=str(e))])
            return request.data.decode('utf-8') or request.args
        return request.args

    def parse_arguments(self, request):
        try:
            arguments = self.get_arguments(request)
            return sepiida.parsing.parse_arguments(arguments)
        except sepiida.parsing.ParseErrors as parse_errors:
            errors = [Error(error_code=e.error_code, title=str(e), status_code=400) for e in parse_errors.errors]
            raise APIException(errors=errors)

    def process_arguments(self, arguments):
        if not self.SIGNATURE:
            return
        self.SIGNATURE.process_arguments(arguments)
        self.fields = arguments.fields.values if arguments.fields else []
        self.filters = {}
        for _filter in arguments.filters:
            self.filters.setdefault(_filter.name, []).append(_filter)

    def dispatch_request(self, *args, **kwargs):
        arguments = self.parse_arguments(flask.request) if self.SIGNATURE else sepiida.parsing.RequestArguments([])
        single_resource = kwargs.pop('_single_resource')
        try:
            method = self._map_request_to_method(flask.request, single_resource)
        except AttributeError:
            return flask.make_response('', 404)

        self.process_arguments(arguments)
        try:
            if self.SIGNATURE and flask.request.method in ('PUT', 'POST'):
                decoded = sepiida.serialization.loads(flask.request)
                try:
                    payload = self.SIGNATURE.unpackage(None, decoded, flask.request.method)
                except fields.UnpackageError as e:
                    raise sepiida.errors.APIException(errors=e.errors)
                kwargs['payload'] = payload

            my_exceptions = tuple([handler.exception_class for handler in self.get_error_handlers()])
            try:
                inspect.getcallargs(method, *args, **kwargs) # pylint: disable=deprecated-method
            except TypeError:
                title = ("You attempted to request a {method} on {path}. "
                    "You must call {method} on a specific resource rather "
                    "than the collection. You can do this by requesting a URL "
                    "of the pattern {path}<uuid>/").format(method=flask.request.method, path=flask.request.path)
                raise Error('invalid-method', title, status_code=405)

            try:
                response = method(*args, **kwargs)
            except my_exceptions as e: # pylint: disable=catching-non-exception
                response = self._convert_exception(e)

            response = self._package_response(flask.request, method, response, arguments.fields)

            return response
        except Error as e:
            raise APIException(errors=[e])

    def options(self, _id=None, uuid=None): # pylint: disable=unused-argument
        if self.SIGNATURE is None:
            return '', 204, {}
        description = {
            'signature': self.SIGNATURE.get_options_description(),
        }
        return flask.make_response(json.dumps(description), 200, {})


NAME_EXTRACTOR = re.compile(r"<class '(?P<name>.+)'>")
def _extract_name(resource):
    # I can't for the life of me figure out how to get the classes'
    # definition heirarchy programmatically. So we'll cheat. Badly.
    # forgive me.
    base = repr(resource)
    match = NAME_EXTRACTOR.match(base)
    if not match:
        raise Exception("Our name extractor doesn't work right and couldn't match '{}'".format(base))
    return match.group('name')

def has_collection_methods(resource):
    return any([hasattr(resource, method) for method in ('delete', 'list', 'post')])

def has_single_methods(resource):
    return any([hasattr(resource, method) for method in ('get', 'put', 'patch', 'delete')])

def get_collection_methods(resource):
    methods = ['HEAD', 'OPTIONS']
    if hasattr(resource, 'post'):
        methods.append('POST')
    if hasattr(resource, 'list'):
        methods.append('GET')
    if hasattr(resource, 'delete'):
        methods.append('DELETE')
    return methods

def get_single_methods(resource):
    methods = ['HEAD', 'OPTIONS']
    for method in ('PUT', 'GET', 'PATCH', 'DELETE'):
        if hasattr(resource, method.lower()):
            methods.append(method)
    return methods

def _check_error_handlers(resource):
    for errorhandler in resource.ERRORS:
        if isinstance(errorhandler, tuple):
            LOGGER.warning((
                "The error handler %s for %s is a tuple. It should be converted to a sepiida.error.Specification. "
                " Until then, it will be ignored"), errorhandler, resource
            )

def add_resource(app, resource, endpoint=None):
    setattr(app, 'endpoints', getattr(app, 'endpoints', {}))

    # This check is to help people transitioning from sepiida pre-7. Remove in sepiida 8
    _check_error_handlers(resource)

    name = _extract_name(resource) if endpoint is None else endpoint
    if resource.ENDPOINT is None:
        raise Exception("You must specify an ENDPOINT for {}".format(resource))
    for public_method in resource.PUBLIC_METHODS:
        if public_method not in ['delete', 'get', 'list', 'post', 'put']:
            raise Exception(
                "Method {} is not supported as PUBLIC_METHOD. "
                "Supported methods are ['delete', 'get', 'list', 'post', 'put']."
                .format(public_method))
        sepiida.session.public_endpoint(app, name + '.' + public_method)

    view_func = resource.as_view(name)
    app.endpoints[name] = resource
    if has_collection_methods(resource):
        collection_methods = get_collection_methods(resource)
        app.add_url_rule(resource.ENDPOINT,
            name,
            view_func,
            methods=collection_methods,
            defaults={'_single_resource': False})

    if has_single_methods(resource):
        single_methods = get_single_methods(resource)
        app.add_url_rule(resource.endpoint_with_uuid(),
            name,
            view_func,
            methods=single_methods,
            defaults={'_single_resource': True})
        app.add_url_rule(resource.endpoint_with_id(),
            name,
            view_func,
            methods=single_methods,
            defaults={'_single_resource': True})



import logging.config


def setup_logging(sentry_dsn=None):
    configuration = {
        'version'   : 1,
        'disable_existing_loggers'  : False,
        'formatters'                : {
            'standard'              : {
                'format'            : '[%(asctime)s] %(levelname)s pid:%(process)d %(name)s:%(lineno)d %(message)s',
                'dateformat'        : '%d/%b/%Y:%H:%M:%S %z',
            },
        },
        'handlers'                  : {
            'console'               : {
                'level'             : 'DEBUG',
                'class'             : 'logging.StreamHandler',
                'formatter'         : 'standard',
            },
        },
        'loggers'                   : {
            ''                      : {
                'handlers'          : ['console'],
                'level'             : 'DEBUG',
                'propogate'         : True,
            },
        },
    }
    if sentry_dsn:
        configuration['handlers']['sentry'] = {
            'level' : 'ERROR',
            'class' : 'raven.handlers.logging.SentryHandler',
            'dsn'   : sentry_dsn,
        }
        configuration['loggers']['']['handlers'].append('sentry')
    logging.config.dictConfig(configuration)



import re
from datetime import datetime


class ParseError(Exception):
    def __init__(self, error_code, message):
        super().__init__(message)
        self.error_code = error_code

class ParseErrors(Exception):
    def __init__(self, errors):
        super().__init__()
        self.errors = errors

class RequestArguments(): # pylint: disable=too-few-public-methods
    def __init__(self, arguments):
        self.fields = [arg for arg in arguments if isinstance(arg, FieldsArgument)]
        assert len(self.fields) < 2, "Should not end up with multiple fields"
        self.fields = self.fields[0] if self.fields else None
        self.filters = [arg for arg in arguments if isinstance(arg, FilterArgument)]

class RequestArgument(): # pylint: disable=too-few-public-methods
    def __init__(self, queryarg, values):
        self.queryarg = queryarg
        self.values   = values

class FilterArgument(RequestArgument): # pylint: disable=too-few-public-methods
    EQ  = '='
    GT  = '>'
    GTE = '>='
    LT  = '<'
    LTE = '<='
    def __init__(self, queryarg, values):
        super().__init__(queryarg, values)
        if len(values) > 1:
            raise ParseError(
                error_code='filter-property-specified-more-than-once',
                message=("Your requested filter property, '{}', was specified more than once."
                    " We don't currently support that").format(queryarg))

        argname = queryarg[6:]
        if not argname:
            raise ParseError(
                error_code='empty-filter-name',
                message=("You requested a filter with no name."
                    " Filter requests should be of the form 'filter[property]'"))
        if argname[-1] == ']':
            self.operation = self.EQ
            self.handle_standard_queryarg(argname, values)
        elif argname[-2:] == ']<':
            self.operation = self.LTE
            self.handle_standard_queryarg(argname, values)
        elif argname[-2:] == ']>':
            self.operation = self.GTE
            self.handle_standard_queryarg(argname, values)
        elif not re.match(r'\[\w+\]', argname):
            raise ParseError(
                error_code='invalid-filter-request',
                message=("You requested a filter 'filter{}'. Filter requests should be of the form 'filter[property]'".format(argname)))
        else:
            self.handle_nonstandard_queryarg(argname, values)

        if not self.values:
            raise ParseError(
                error_code='empty-filter-value',
                message=("You requested a filter with no values. This won't match anything, ever"))

    def handle_nonstandard_queryarg(self, argname, values):
        if values and not '' in values:
            raise ParseError(
                error_code='invalid-filter-request',
                message=(("You requested a filter 'filter{}={}' with an equal in the value. "
                    "Filter requests should be of the form 'filter[property]=value'").format(argname, values[0])))

        match = re.match(r'\[(?P<name>\w+)\](?P<operation>\<|\>)(?P<values>[ ,\w,-]+)$', argname)
        if not match:
            raise ParseError(
                error_code='invalid-filter-request',
                message=("You requested a filter 'filter{}'. Filter requests should be of the form 'filter[property]'".format(argname)))
        if match.group('operation') == '<':
            self.operation = self.LT
        elif match.group('operation') == '>':
            self.operation = self.GT
        else:
            raise ParseError(error_code='programmer-error', message="Programming assertion failed. Tell Authentise support")

        self.name = match.group('name')
        self.values = match.group('values').split(',')

    def handle_standard_queryarg(self, argname, values):
        try:
            self.values = values[0].split(',')
        except IndexError:
            self.values = []

        if not argname[0] == '[':
            raise ParseError(
                error_code='invalid-filter-request',
                message=("You requested a filter 'filter{}'."
                    " Filter requests should be of the form 'filter[property]'").format(argname))

        self.name = argname[1:(-1 if self.operation == self.EQ else -2)]
        if not self.name:
            raise ParseError(
                error_code='empty-filter-name',
                message=("You requested a filter 'filter{}' with no name."
                    " Filter requests should be of the form 'filter[property]'").format(argname))

    def __str__(self):
        formatted_values = [value.isoformat() if isinstance(value, datetime) else str(value) for value in self.values]
        return "FilterArgument({}={})".format(self.name, ','.join(formatted_values))

    def __repr__(self):
        return str(self)

class FieldsArgument(RequestArgument): # pylint: disable=too-few-public-methods
    def __init__(self, queryarg, values):
        super().__init__(queryarg, values)
        if len(values) > 1:
            suggestion = "fields=[{}]".format(','.join(sorted(values)))
            raise ParseError(
                    error_code='fields-property-specified-more-than-once',
                    message=("You specified fields more than once. We don't support that. "
                             "Please combine multiple fields using a comma such as {}").format(suggestion))

        _fields = values[0]
        if not (_fields[0] == '[' and _fields[-1] == ']'):
            raise ParseError(
                error_code='invalid-fields-query',
                message=("You specified a fields value of '{}'. Fields must be specified "
                        "as a comma-separated list of fields enclosed in '[' and ']'")
                        .format(_fields)
                )
        _fields = _fields[1:-1]
        _fields = _fields.split(',')
        self.values = [f.strip() for f in _fields]

ARGUMENT_PATTERNS = {
    re.compile(r'filter')   : FilterArgument,
    re.compile(r'fields')   : FieldsArgument,
}
def _get_matching_pattern(queryarg):
    for pattern, cls in ARGUMENT_PATTERNS.items():
        match = pattern.match(queryarg)
        if match:
            return cls
    raise ParseError('unrecognized-query-argument', 'The query argument {} is not recognized by this endpoint'.format(queryarg))

def parse_arguments(queryargs):
    arguments = []
    errors = []
    for queryarg in queryargs.keys():
        try:
            cls = _get_matching_pattern(queryarg)

            values = queryargs.getlist(queryarg)
            argument = cls(queryarg, values)
            arguments.append(argument)
        except ParseError as e:
            errors.append(e)
    if errors:
        raise ParseErrors(errors=errors)
    return RequestArguments(arguments)



import collections
import inspect
import json
import logging

import accept_types
import flask
import yaml

import sepiida.errors
import sepiida.session

LOGGER = logging.getLogger(__name__)

def schema(resource, method):
    if not resource.SIGNATURE:
        return
    my_schema = {
        "$schema": "http://json-schema.org/draft-03/schema",
    }
    my_schema.update(resource.SIGNATURE.schema(method=method))
    return json.dumps(my_schema, default=str)

def example(resource):
    if not resource.SIGNATURE:
        return
    return json.dumps(resource.SIGNATURE.example)

def responses(resource, method):
    if method in ('LIST', 'GET'):
        my_schema = schema(resource, 'POST')
        my_example = example(resource)
        result = {'200': {
            'body'  : {
                'application/json'  : {
                    'example'       : my_example,
                    'schema'        : my_schema,
                },
            },
            'description'   : 'The request was successful',
        }}
    elif method == 'POST':
        result = {'204': {
            'description'   : 'The request was successful'
        }}
    else:
        raise Exception("Unrecognized method {}".format(method))

    descriptions = collections.defaultdict(lambda: [])
    for spec in resource.ERRORS:
        if spec.docs:
            description = spec.docs
        elif spec.error_code and spec.title:
            description = "`{}`: {}".format(spec.error_code, spec.title)
        elif spec.title:
            description = spec.title
        elif spec.error_code:
            description = "`{}`".format(spec.error_code)
        else:
            description = "{} raised".format(spec.exception_class.__name__)
        descriptions[str(spec.status_code or 400)].append(description)
    for k, v in descriptions.items():
        result[k] = {'description'   : '\n'.join(v)}
    return result

def method_description(resource, method):
    handler = getattr(resource, method.lower(), None)
    if not handler:
        return
    docstring = inspect.getdoc(handler)
    my_example = example(resource)
    result = {
        'description'   : docstring,
        'responses'     : responses(resource, method),
        'body'          : {
            'application/json'  : {
            }
        }
    }
    if my_example:
        result['body']['application/json']['example'] = my_example
    my_schema = schema(resource, 'POST')
    if my_schema:
        result['body']['application/json']['schema'] = my_schema
    return result

def collection_descriptions(resource):
    description = {
        'description'   : inspect.getdoc(resource),
    }
    if hasattr(resource, 'list'):
        description['get'] = method_description(resource, method='LIST')
    if hasattr(resource, 'post'):
        description['post'] = method_description(resource, method='POST')
    return description

def single_descriptions(resource):
    return {
        'description'   : inspect.getdoc(resource),
        'uriParameters' : {
            'uuid'      : {
                'description'   : "The UUID of the resource",
            }
        },
        'get'   : method_description(resource, method='GET'),
    }

def get_description(resource):
    LOGGER.debug("Getting description of %s", resource)
    descriptions = {}
    if sepiida.endpoints.has_collection_methods(resource):
        descriptions[resource.ENDPOINT] = collection_descriptions(resource)
        LOGGER.debug("Added %s: %s", resource.ENDPOINT, descriptions[resource.ENDPOINT])
    if sepiida.endpoints.has_single_methods(resource):
        uri = resource.ENDPOINT + "{uuid}/"
        descriptions[uri] = single_descriptions(resource)
        LOGGER.debug("Added %s: %s", uri, descriptions[uri])
    return descriptions

def handle_raml_request(name, baseURL, documentation, version):
    descriptions = {}
    for resource in flask.current_app.endpoints.values():
        descriptions.update(get_description(resource))
    descriptions.update({
        'title'         : name,
        'documentation' : documentation or [{'content': '', 'title': ''}],
        'version'       : version,
        'baseUri'       : baseURL,
        'mediaType'     : 'application/json',
    })
    content = yaml.dump(descriptions, default_style='|')
    return flask.make_response('#%RAML 0.8\n{}'.format(content), 200, {'Content-Type': 'application/raml+yaml'})

def handle_request(name, baseURL, documentation, version):
    accept = flask.request.headers.get('Accept', 'text/raml')
    matched_type = accept_types.get_best_match(accept, ('application/raml+yaml', 'text/raml'))
    LOGGER.debug("Handling OPTIONS request for %s", matched_type)
    if matched_type in ('text/raml', 'application/raml+yaml'):
        return handle_raml_request(name, baseURL, documentation, version)
    error = sepiida.errors.APIException(errors=[sepiida.errors.UnableToSerialize()])
    return flask.make_response(error.get_body(), error.code, error.get_headers())

def enable(app, name, baseURL, version, documentation=None):
    app.add_url_rule('/', 'root', handle_request, methods=['OPTIONS'], defaults={
        'baseURL'       : baseURL,
        'documentation' : documentation,
        'name'          : name,
        'version'       : version
    })
    sepiida.session.public_endpoint(app, 'root.options')



import logging
import re
from urllib import parse

import flask
import requests
import requests.auth
import sepiida.context
import sepiida.errors
import sepiida.session

LOGGER = logging.getLogger(__name__)

def user_session(username=None, password=None, session=None, user=None):
    if not any([username, password, session, user]):
        if flask.has_request_context():
            credentials = sepiida.context.extract_credentials(flask.request)
            password    = credentials['password']
            session     = credentials['session']
            user        = credentials['user']
            username    = credentials['username']
        elif sepiida.backend.has_task_context():
            credentials = sepiida.backend.task_credentials()
            password    = credentials['password']
            session     = credentials['session']
            user        = credentials['user']
            username    = credentials['username']
        else:
            msg = ("You have attempted to create a user session without having a flask request context or a backend task context. "
                   "You must therefore supply your own credentials as parameters")
            raise Exception(msg)

    _user_session = requests.Session()
    if username and password:
        LOGGER.debug("Using credentials %s %s for user session", username, password)
        _user_session.auth = requests.auth.HTTPBasicAuth(username, password)
    elif session:
        LOGGER.debug("Passing through session token")
        _user_session.cookies['session'] = session
    else:
        LOGGER.warning("User session requested, but I have no session data to use")
    return _user_session

class PrivilegedAuth(requests.auth.HTTPBasicAuth): # pylint: disable=too-few-public-methods
    def __call__(self, prepared_request):
        prepared_request = super(self.__class__, self).__call__(prepared_request)
        parts = parse.urlsplit(prepared_request.url)
        domains = flask.current_app.config.get('SEPIIDA_INTERNAL_DOMAINS', [])
        domains = '|'.join([re.escape(domain) for domain in domains])
        domains_regex = r'(.*)?\.({0})$'.format(domains)
        if not parts.scheme == 'https':
            raise InvalidPrivilegedScheme()
        if not re.fullmatch(domains_regex, parts.netloc):
            raise InvalidPrivilegedDomain()

        return prepared_request

def privileged_session(api_token=None):
    if not api_token:
        api_token = flask.current_app.config['API_TOKEN']

    _session = requests.Session()
    _session.auth = PrivilegedAuth('api', api_token)
    return _session

class InvalidPrivilegedDomain(sepiida.errors.Error):
    def __init__(self):
        super(
            self.__class__,
            self,
        ).__init__(
            "Privileged domain must be trusted",
            "invalid-privileged-domain"
        )

class InvalidPrivilegedScheme(sepiida.errors.Error):
    def __init__(self):
        super(
            self.__class__,
            self,
        ).__init__(
            "Privileged request requires https",
            "invalid-privileged-sheme"
        )



import contextlib
import logging

import flask
import requests

import sepiida.routing
from sepiida.requests import privileged_session, user_session

LOGGER = logging.getLogger(__name__)

CONFIG = {
  'PAO_ROOT': None,
}
def configure(root):
    CONFIG['PAO_ROOT'] = root

class PermissionsRequestError(Exception):
    pass

class PermissionsConnectionError(Exception):
    pass

class PermissionsPaoRootError(Exception):
    pass

def _send_request(http_method, uri=None, query=None, payload=None, timeout=10, privileged=False):
    privileged = is_privileged() or privileged
    pao_root = CONFIG.get('PAO_ROOT')
    if not pao_root:
        raise PermissionsPaoRootError(
            "You must include a value for PAO_ROOT in your app config,"
            "such as https://pao.service, before querying the permissions system"
        )

    url = uri or '{}/permissions/'.format(pao_root)

    session = privileged_session() if privileged else user_session()
    try:
        if query:
            response = session.request(http_method, url, data=query, timeout=timeout)
        else:
            response = session.request(http_method, url, json=payload, timeout=timeout)
    except (requests.exceptions.ReadTimeout, requests.exceptions.ConnectionError) as e:
        raise PermissionsConnectionError(e)

    if not response.ok:
        LOGGER.error('Error sending request %s %s with payload %s: %s %s',
            http_method,
            url,
            payload,
            response.status_code,
            response.text,
        )
        raise PermissionsRequestError(
            "Failed Request, Response: {}, {}, Payload: {}".format(response.status_code, response.text, payload)
        )
    return response

def _payload(object_, holder, right, namespace):
    return {
        'object'    : object_,
        'holder'    : holder,
        'right'     : right,
        'namespace' : namespace,
    }

def has_right(object_, right, namespace):
    permissions = search(objects=[object_], rights=[right], namespace=namespace)
    return len(permissions) > 0

def has_any(object_=None, holder=None, right=None, namespace=None):
    permissions = search(
        objects     = [object_] if object_ else None,
        holders     = [holder] if holder else None,
        rights      = [right] if right else None,
        namespace   = namespace if namespace else None,
    )
    return len(permissions) > 0

def get(uri=None, privileged=False):
    http_method = 'GET'
    response = _send_request(http_method, uri=uri, privileged=privileged)
    return response.json()

@contextlib.contextmanager
def always_privileged():
    is_privileged.nest += 1
    yield
    is_privileged.nest -= 1

def is_privileged():
    return is_privileged.nest > 0
is_privileged.nest = 0

def search(objects=None, holders=None, rights=None, namespace=None, privileged=False):
    http_method = 'GET'
    filters = []
    if objects:
        filters.append("filter[object]={}".format(','.join(objects)))
    if holders:
        filters.append("filter[holder]={}".format(','.join(holders)))
    if rights:
        filters.append("filter[right]={}".format(','.join(rights)))
    if namespace:
        filters.append("filter[namespace]={}".format(namespace))
    filters_qs = "&".join(filters)
    response = _send_request(http_method, query=filters_qs, privileged=privileged)
    return response.json()['resources']

def update_filters(filters, namespace, resource_name, rights=None, app=None):
    """
    Update the provided filters so that permissions for the current user are enforced when querying the database
    The idea here is that pretty much all platform layers will include the ability to filter
    on a particular uuid for a resource. We enforce permissions by limiting their
    query in the database to a list of rows they are allowed to see.
    If the filter already limits to a subset of rows we take the union of the set
    they can see and the set they requested and update the filter accordingly
    This is designed to make it possible to apply filters in a single line
    """
    if is_privileged():
        LOGGER.debug("Refusing to constrain filters because we are currently in an always_privileged context")
        return filters
    app = app or flask.current_app
    permissions = search(rights=rights, namespace=namespace, privileged=False)
    resources = [permission['object'] for permission in permissions]
    parameters = [sepiida.routing.extract_parameters(app, 'GET', resource) for resource in resources]
    resource_uuids = {params['uuid'] for endpoint, params in parameters if endpoint == resource_name}
    requested_uuids = filters.get('uuid', None)
    common_uuids = resource_uuids.intersection(requested_uuids) if requested_uuids else resource_uuids
    filters['uuid'] = common_uuids
    return filters

def create(object_, holder, right, namespace, privileged=False):
    http_method = 'POST'
    payload = _payload(object_, holder, right, namespace)
    response = _send_request(http_method, payload=payload, privileged=privileged)
    return response.headers['Location']

def update(uri, object_, holder, right, namespace, privileged=False):
    http_method = 'PUT'
    payload = _payload(object_, holder, right, namespace)
    response = _send_request(http_method, uri=uri, payload=payload, privileged=privileged)
    assert response.status_code == 204

def delete(uri, privileged=False):
    http_method = 'DELETE'
    response = _send_request(http_method, uri=uri, privileged=privileged)
    assert response.status_code == 204

def delete_all(objects=None, holders=None, rights=None, namespace=None, privileged=False):
    permissions = search(
        objects   = objects,
        holders   = holders,
        rights    = rights,
        namespace = namespace
    )
    for permission in permissions:
        delete(permission['uri'], privileged=privileged)



import logging

import flask
from raven.contrib.flask import Sentry


class SepiidaSentry(Sentry):
    def get_user_info(self, request):
        return flask.g.current_user

def add_sentry_support(application, sentry_dsn):
    raven = SepiidaSentry(application, dsn=sentry_dsn, logging=True, level=logging.ERROR)
    application.config.setdefault('RAVEN', raven)
    logging.getLogger('webargs.flaskparser').setLevel(logging.CRITICAL)



import hashlib
import json as JSON
import logging
from datetime import datetime, timedelta

import jwt
from flask import current_app, g, request

from sepiida.errors import APIException, Error
from sepiida.requests import privileged_session

LOGGER = logging.getLogger(__name__)

ALGORITHM = 'RS512'

def register_jwt_handlers(app, secret_key=None, secret_key_path=None):
    app.config['SEPIIDA_JWT_KEY'] = secret_key or read_secret_key(secret_key_path)
    app.before_request(jwt_handler)

def jwt_handler():
    _jwt = request.headers.get('jwt')
    if _jwt:
        try:
            g.jwt = validate(_jwt)
        except Error as e:
            raise APIException(errors=[e])

def request_jwt():
    return g.get('jwt', None)

def validate(token, payload=None, audience=None):
    iss = decode_iss(token)
    session = privileged_session()
    response = session.get(iss, timeout=5)
    if not response.ok:
        LOGGER.warning('Issuer (%s) request failed with status code %s: %s', iss, response.status_code, response.text)
        raise InvalidIssuerError("jwt-invalid-issuer", "Invalid issuer")

    response_json = response.json()
    if not response_json:
        raise InvalidIssuerError("jwt-invalid-issuer-response", "Invalid issuer response")

    secret = response_json.get('secret')
    if not secret:
        LOGGER.warning('Issuer (%s) request failed with missing secret token', iss)
        raise InvalidIssuerError("jwt-invalid-issuer-secret", "Invalid issuer secret")

    if not audience:
        audience = request.url_root

    if not payload:
        payload = request.get_data()

    return decode(token, cert=secret, payload=payload, audience=audience)

def decode(token, cert=None, payload=None, verify=True, audience=None, issuer=None):
    options = {
        "require_exp": True,
        "require_iat": True,
        "require_nbf": True,
    }

    try:
        body = jwt.decode(token, cert, options=options, verify=verify, audience=audience, issuer=issuer, leeway=10)
    except jwt.exceptions.MissingRequiredClaimError as ex:
        LOGGER.warning(ex)
        raise MissingClaimError("jwt-missing-claim", str(ex))
    except jwt.exceptions.InvalidIssuerError as ex:
        LOGGER.warning(ex)
        raise InvalidIssuerError("jwt-invalid-issuer", str(ex))
    except jwt.exceptions.InvalidAudienceError as ex:
        LOGGER.warning(ex)
        raise InvalidAudienceError("jwt-invalid-audience", str(ex))
    except jwt.exceptions.ExpiredSignatureError as ex:
        LOGGER.warning(ex)
        raise ExpiredSignatureError("jwt-signature-expired", str(ex))
    except jwt.exceptions.DecodeError as ex:
        LOGGER.warning(ex)
        raise DecodeError("jwt-decode-error", str(ex))

    if "exp" in body:
        body["exp"] = datetime.utcfromtimestamp(body["exp"])

    if "iat" in body:
        body["iat"] = datetime.utcfromtimestamp(body["iat"])

    if "nbf" in body:
        body["nbf"] = datetime.utcfromtimestamp(body["nbf"])

    if not verify:
        return body

    if "bdy" in body:
        bdy = body["bdy"]
        expected_bdy = _encode_payload(payload)
        if bdy != expected_bdy:
            LOGGER.warning('Invalid bdy claim: bdy "%s" does not match payload "%s"', bdy, expected_bdy)
            raise InvalidBodyError("jwt-invalid-body", 'Invalid Body')

    return body

def encode(key=None, **kwargs):
    utcnow = datetime.utcnow()
    key = key if key else current_app.config.get('SEPIIDA_JWT_KEY')
    body = kwargs.copy()

    if "issuer" in body:
        body["iss"] = body.pop("issuer")

    if "subject" in body:
        body["sub"] = body.pop("subject")

    if "audience" in body:
        body["aud"] = body.pop("audience")

    if "exp" not in body:
        body["exp"] = utcnow + timedelta(seconds=900)

    if "iat" not in body:
        body["iat"] = utcnow

    if "nbf" not in body:
        body["nbf"] = utcnow

    payload = body.pop("payload", None)
    if payload:
        body["bdy"] = _encode_payload(payload)

    return jwt.encode(body, key, algorithm=ALGORITHM)

def decode_iss(token):
    body = decode(token, None, None, verify=False)
    iss = body.get("iss")
    if not iss:
        raise MissingClaimError("jwt-missing-cliam", 'Token is missing the "iss" claim')
    return iss

def read_secret_key(path=None):
    with open(path or "/etc/authentise/authentise.jwt.key.pem") as f:
        return f.read()

def _encode_payload(payload):
    if isinstance(payload, dict):
        payload = JSON.dumps(payload).encode()
    elif isinstance(payload, str):
        payload = payload.encode()
    return hashlib.sha256(payload).hexdigest()

class DecodeError(Error):
    pass

class InvalidClaimError(Error):
    pass

class MissingClaimError(Error):
    pass

class InvalidIssuerError(Error):
    pass

class InvalidAudienceError(Error):
    pass

class InvalidBodyError(Error):
    pass

class ExpiredSignatureError(Error):
    pass



import inspect
import json
import logging
import math
import urllib.parse
import uuid
from datetime import date, datetime

import arrow
import flask

import sepiida.routing
from sepiida.errors import APIException, AppError, Error

LOGGER = logging.getLogger(__name__)

class UnpackageError(Exception):
    def __init__(self, errors):
        super().__init__()
        assert isinstance(errors, list)
        self.errors = errors

class PackageError(Exception):
    def __init__(self, errors):
        super().__init__()
        assert isinstance(errors, list)
        self.errors = errors
    def __str__(self):
        return "\n".join(self.errors)

TYPES_TO_NAME = {
    date        : 'a date',
    datetime    : 'a datetime',
    dict        : 'an object',
    list        : 'an array',
    int         : 'a number',
    float       : 'a number',
    bool        : 'a boolean',
    str         : 'a string',
    type(None)  : 'null',
    uuid.UUID   : 'a uuid',
    'URI'       : 'a uri',
    'URL'       : 'a URL',
}
def _unpackage_error_message(name, expected_type, provided_data):
    provided_type_name = TYPES_TO_NAME[type(provided_data)]
    expected_type_name = TYPES_TO_NAME[expected_type]
    return "The property '{}' must be {}. You provided {}".format(name, expected_type_name, provided_type_name)

def _raise_validation_error(title):
    errors = [Error('request-validation-error', title)]
    raise UnpackageError(errors)

def _raise_validation_type_error(name, expected_type, provided_data):
    title = _unpackage_error_message(name, expected_type, provided_data)
    errors = [Error('type-validation-error', title)]
    raise UnpackageError(errors)

def _maybe_raise_all_sub_errors(errors):
    if not errors:
        return

    sub_errors = []
    for e in errors:
        sub_errors += e.errors
    raise UnpackageError(errors=sub_errors)

class Unspecified(): # pylint: disable=too-few-public-methods
    pass

class Field(): # pylint: disable=too-many-instance-attributes
    TYPE_NAME = 'raw-field'
    def __init__(self, # pylint: disable=too-many-arguments
            choices    = None,
            default    = Unspecified,
            docs       = None,
            example    = Unspecified,
            filterable = True,
            methods    = None,
            nullable   = False,
            optional   = False,
            rename     = None,
        ):
        self.choices = choices
        if default != Unspecified:
            self.default = default
        self.docs           = docs
        self.filterable     = filterable
        self.methods        = methods if methods is not None else ['GET', 'POST', 'PUT']
        self.nullable       = nullable
        self.optional       = optional
        self.rename         = rename

        self._generate_example(example, choices)

    def _check_choices(self, name, data):
        if self.choices is None:
            return

        if data not in self.choices:
            choices = ["'{}'".format(choice) for choice in self.choices]
            choices = ", ".join(choices)
            error = Error(
                'property-value-not-in-allowed-values',
                "The property '{}' must be one of {}. You provided '{}'".format(name, choices, data)
            )
            raise UnpackageError([error])

    def _generate_example(self, example, choices):
        if example != Unspecified:
            self.example = example
        elif choices:
            self.example = '|'.join([str(choice) for choice in choices])

    def package(self, data):
        raise NotImplementedError

    def unpackage(self, name, package, method):
        raise NotImplementedError

    def get_options_description(self):
        description = {
            'type'          : self.TYPE_NAME,
            'choices'       : self.choices,
            'filterable'    : self.filterable,
            'methods'       : self.methods,
            'optional'      : self.optional
        }
        if hasattr(self, 'default'):
            description['default'] = self.default
        return description

    def get_filter_value(self, name, value): # pylint: disable=unused-argument,no-self-use
        return value

    def schema(self, method='POST'):
        my_schema = {
            'description'   : self.docs,
            'filterable'    : self.filterable,
            'nullable'      : self.nullable,
            'type'          : self.TYPE_NAME
        }
        if method == 'GET':
            my_schema['optional'] = self.optional
        if method == 'POST' and hasattr(self, 'default'):
            my_schema['default'] = self.default
        if self.choices:
            my_schema['enum'] = self.choices
        if hasattr(self, 'example'):
            my_schema['example'] = json.dumps(self.example)
        return my_schema


class SignatureField(Field): # pylint: disable=abstract-method
    def __init__(self, s=None, **kwargs):
        super().__init__(**kwargs)
        self.signature = s
        root_signature = self.signature
        while isinstance(root_signature, SignatureField):
            root_signature = root_signature.signature

        if root_signature is None:
            return

        try:
            for k, v in root_signature.items():
                if inspect.isclass(v):
                    raise Exception(("You have provided a class '{}' as part of the signature for {}."
                        "Please specify the signature using instances of the class, not the class itself".format(v, k)))
        except AttributeError:
            pass

        if isinstance(root_signature, dict):
            fieldnames = set(root_signature.keys())
            for key, field in root_signature.items():
                if field.rename and field.rename in fieldnames:
                    raise Exception((
                        "You cannot rename field {} to {} because another "
                        "field is already using that name").format(key, field.rename))

        try:
            self._fieldnames = ', '.join(["'{}'".format(name) for name in sorted(root_signature.keys())])
            _filterable = [name for name, field in root_signature.items() if field.filterable]
            self._filternames = ', '.join(["'{}'".format(name) for name in sorted(_filterable)])
        except AttributeError:
            self._fieldnames = None


    def _process_against_signature(self, arguments):
        errors = []
        if arguments.fields and not self.signature:
            errors.append(Error(
                status_code=400,
                error_code='unable-to-specify-fields',
                title=("You specified return fields. This resource cannot return a subset of its fields"),
            ))

        if arguments.filters and not self.signature:
            errors.append(Error(
                status_code=400,
                error_code='unable-to-specify-filters',
                title=("You specified filters. This resource cannot be filtered"),
            ))
        return errors

    def _process_fields(self, arguments):
        if not self.signature:
            return []

        errors = []
        for field in (arguments.fields.values if arguments.fields else []):
            if field not in self.signature:
                msg = ("You specified the resource return '{}' but it is not a valid field for this resource. "
                       "Valid fields include {}").format(field, self._fieldnames)
                errors.append(Error('invalid-field-specified', msg))
        return errors

    def _process_filters(self, arguments):
        if not self.signature:
            return []

        errors = []
        for _filter in arguments.filters:
            try:
                field = self.signature[_filter.name]
            except KeyError:
                if self.signature:
                    title = ("Your requested filter property, '{}', is not a valid property for this endpoint. "
                            "Please choose one of {}".format(_filter.name, self._filternames))
                else:
                    title = "You requested a filter on property '{}'. This endpoint does not allow filtering".format(_filter.name)
                errors.append(Error('invalid-filter-property', title))
                continue

            if not field.filterable:
                title = ("Your requested filter property, '{}', is not a valid property for this endpoint. "
                         "Please choose one of {}").format(_filter.name, self._filternames)

                errors.append(Error(
                    status_code = 400,
                    error_code  = 'invalid-filter-property',
                    title       = title,
                ))
                continue

            new_values = []
            for value in _filter.values:
                try:
                    new_value = field.get_filter_value(_filter.name, value)
                    if field.choices and new_value not in field.choices:
                        title = ("You requested a filter on '{name}' to the value '{value}'. "
                                 "The property '{name}' only permits values from a particular "
                                 "set of options. Please use one of the following options for "
                                 "your filter: {values}").format(name=_filter.name, value=value, values=', '.join(field.choices))
                        errors.append(Error(
                            status_code = 400,
                            error_code  = 'invalid-filter-value',
                            title       = title,
                        ))
                        continue
                    new_values.append(new_value)
                except ValueError:
                    title = ("You requested a filter on '{name}' to the value '{value}'. "
                            "The property '{name}' is a {type} and so filters on it must "
                            "also be {type}").format(name=_filter.name, value=value, type=field.TYPE_NAME)
                    errors.append(Error(
                        status_code = 400,
                        error_code  = 'invalid-filter-value',
                        title       = title,
                    ))
                except AppError as e:
                    errors.append(Error(
                        status_code = 500,
                        error_code  = 'app-error',
                        title       = str(e)
                    ))
            _filter.values = new_values
        return errors

    def process_arguments(self, arguments):
        errors = self._process_against_signature(arguments)
        errors += self._process_fields(arguments)
        errors += self._process_filters(arguments)

        if errors:
            raise APIException(errors=errors)

class Object(SignatureField):
    TYPE_NAME = 'object'
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.signature:
            self.example = {k: v.example for k, v in self.signature.items() if hasattr(v, 'example')}
        else:
            self.example = None

    def package(self, response):
        package = {}
        if response is None:
            return None

        if self.signature is None:
            return dict(response)

        if response is None:
            return None

        errors = []
        for k, v in self.signature.items():
            try:
                if 'GET' in v.methods:
                    try:
                        newkey = k if v.rename is None else v.rename
                        package[k] = v.package(response[newkey])
                    except (TypeError, ValueError) as e:
                        errors.append("Programmer failed to provide the correct type for {}: {}".format(k, e))
            except KeyError:
                if not v.optional:
                    errors.append("Programmer failed to provide {}, which is a required value to return for this resource.".format(k))
        if errors:
            raise PackageError(errors=errors)
        return package

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, dict):
            raise _raise_validation_type_error(name, dict, data)

        self._check_choices(name, data)

        if self.signature is None:
            return data

        package = {}
        errors = []
        for k, v in data.items():
            try:
                sub_name = "{}{}".format(name + "." if name else '', k)
                try:
                    sub_signature = self.signature[k]
                except KeyError:
                    raise _raise_validation_error("You provided '{}'. It is not a valid property for this resource".format(sub_name))
                value = sub_signature.unpackage(sub_name, v, method)
            except UnpackageError as e:
                errors.append(e)
                value = None
            package[k] = value
        for k, v in self.signature.items():
            sub_name = "{}{}".format(name + "." if name else '', k)
            if method not in v.methods and k in package:
                post_errors = [Error('cannot-set-property', "You provided '{}'. It cannot be directly set via {}".format(sub_name, method))]
                errors.append(UnpackageError(post_errors))
            if method == 'POST' and method in v.methods and k not in package:
                try:
                    package[k] = getattr(v, 'default')
                except AttributeError:
                    post_errors = [Error('missing-parameter',
                        "You did not provide the {} property. It is required when POSTing to this resource".format(sub_name)
                    )]
                    errors.append(UnpackageError(post_errors))
            if v.rename is not None:
                package[v.rename] = package.pop(k)
        _maybe_raise_all_sub_errors(errors)
        return package

    def get_options_description(self):
        base = super().get_options_description()
        if self.signature:
            base['fields'] = {k: v.get_options_description() for k, v in self.signature.items()}
        return base

    def schema(self, method='POST'):
        my_schema = super().schema(method)
        if not self.signature:
            return my_schema
        my_schema.update({
            'properties'    : {k: v.schema(method) for k, v in self.signature.items()},
        })
        if method == 'POST':
            my_schema['required'] = [k for k, v in self.signature.items() if not hasattr(v, 'default')]
        elif method == 'GET':
            my_schema['required'] = [k for k, v in self.signature.items() if not v.optional]
        return my_schema

class Array(SignatureField):
    TYPE_NAME = 'array'
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.example  = [self.signature.example if self.signature and hasattr(self.signature, 'example') else '...']

    def package(self, response):
        if response is None:
            return None

        _convert = (lambda x: self.signature.package(x)) if self.signature else (lambda x: x) # pylint: disable=unnecessary-lambda
        return [_convert(value) for value in response]

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, list):
            raise _raise_validation_type_error(name, list, data)

        self._check_choices(name, data)

        if self.signature is None:
            return data

        package = []
        errors = []
        for i, d in enumerate(data):
            try:
                sub_name = "{}{}".format(name + "." if name else '', i)
                value = self.signature.unpackage(sub_name, d, method)
            except UnpackageError as e:
                errors.append(e)
                value = None
            package.append(value)
        _maybe_raise_all_sub_errors(errors)
        return package

    def schema(self, method='POST'):
        my_schema = super().schema(method)
        if not self.signature:
            return my_schema
        my_schema.update({
            'items'    : self.signature.schema(method=method)
        })
        return my_schema
# Backwards compatibility with previous sepiida versions
ist = Array

class String(Field):
    TYPE_NAME = 'string'
    def package(self, response):
        return str(response) if response is not None else None

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, str):
            raise _raise_validation_type_error(name, str, data)

        self._check_choices(name, data)

        return data

class Integer(Field):
    TYPE_NAME = 'integer'
    def package(self, response):
        return int(response) if response is not None else None

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, int) or isinstance(data, bool):
            raise _raise_validation_type_error(name, int, data)

        self._check_choices(name, data)

        return int(data)

    def get_filter_value(self, name, value):
        parsed = float(value)
        if int(parsed) != parsed:
            raise ValueError("It's a float but not an int")
        return int(parsed)

class Float(Field):
    TYPE_NAME = 'float'
    def package(self, response):
        if isinstance(response, float) and any([math.isnan(response), math.isinf(response)]):
            raise ValueError("Programmer provided '{}' which cannot be serialized correctly".format(response))
        return float(response) if response is not None else None

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        acceptable_types = [isinstance(data, x) for x in (int, float)]
        unacceptable_types = [isinstance(data, bool)]
        if not any(acceptable_types) or any(unacceptable_types):
            raise _raise_validation_type_error(name, float, data)

        self._check_choices(name, data)

        return float(data)

    def get_filter_value(self, name, value):
        return float(value)

class Boolean(Field):
    TYPE_NAME = 'boolean'
    def package(self, response):
        return bool(response) if response is not None else None

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, bool):
            raise _raise_validation_type_error(name, bool, data)

        self._check_choices(name, data)

        return bool(data)

    def get_filter_value(self, name, value):
        lower = value.lower()
        if lower in ('f', 'false') or lower == '0':
            return False
        elif lower in ('t', 'true') or lower == '1':
            return True
        else:
            raise ValueError("{} does not parse to a bool".format(value))

class Date(Field):
    TYPE_NAME = 'iso-8601 date'
    def package(self, response):
        if response is None:
            return None

        if isinstance(response, date):
            return response.isoformat()
        else:
            raise TypeError("{} is not a valid date".format(response))

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, str):
            raise _raise_validation_type_error(name, date, data)

        try:
            return arrow.get(data).date()
        except arrow.parser.ParserError:
            raise _raise_validation_type_error(name, date, data)

    def get_filter_value(self, name, value):
        try:
            return arrow.get(value).date()
        except arrow.parser.ParserError:
            raise ValueError("{} does not parse to a date".format(value))

class Datetime(Field):
    TYPE_NAME = 'iso-8601 datetime'
    def package(self, response):
        if response is None:
            return None

        if isinstance(response, datetime):
            return response.isoformat()
        else:
            raise TypeError("{} is not a valid datetime".format(response))

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, str):
            raise _raise_validation_type_error(name, datetime, data)

        try:
            return arrow.get(data).datetime
        except arrow.parser.ParserError:
            raise _raise_validation_type_error(name, datetime, data)

    def get_filter_value(self, name, value):
        try:
            return arrow.get(value).datetime
        except arrow.parser.ParserError:
            raise ValueError("{} does not parse to a datetime".format(value))

class UUID(Field):
    TYPE_NAME = 'uuid'
    def package(self, response):
        if response is None:
            return None

        if isinstance(response, uuid.UUID):
            return str(response)
        else:
            raise ValueError("{} is not a valid UUID".format(response))

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None

        if not isinstance(data, str):
            raise _raise_validation_type_error(name, uuid.UUID, data)

        try:
            return uuid.UUID(data)
        except ValueError:
            raise _raise_validation_type_error(name, uuid.UUID, data)

    def get_filter_value(self, name, value):
        return uuid.UUID(value)

def _app_has_endpoint(app, endpoint):
    for rule in app.url_map.iter_rules():
        if rule.endpoint == endpoint:
            return True
    return False

class URI(String):
    TYPE_NAME = 'URI'
    def __init__(self, endpoint, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = endpoint

    def package(self, response):
        if response is None:
            return None
        elif isinstance(response, (int, uuid.UUID)):
            return sepiida.routing.uri(self.endpoint, response)
        elif isinstance(response, str):
            endpoint, _ = sepiida.routing.extract_parameters(flask.current_app, method='GET', url=response)
            assert endpoint == self.endpoint, "URI is valid but for the wrong endpoint"
            return response
        else:
            return sepiida.routing.uri(self.endpoint, **response)

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None
        try:
            route, parameters = sepiida.routing.extract_parameters(flask.current_app, method='GET', url=data)
            if not route:
                raise _raise_validation_error("The URI you provided for {} - {} - is not recognized as a URI".format(name, data))
            if route != self.endpoint:
                msg = (
                    "The URI you provided for {property_name} - {data} - "
                    "appears to be a URI for a {bad_route} resource, not a {good_route} resource".format(
                        property_name   = name,
                        data            = data,
                        bad_route       = route,
                        good_route      = self.endpoint,
                ))
                raise _raise_validation_error(msg)
        except AttributeError:
            raise _raise_validation_type_error(name, 'URI', data)
        parameters = parameters or {}
        parameters['uri'] = data
        return parameters

    def get_filter_value(self, name, value):
        _, params = sepiida.routing.extract_parameters(flask.current_app, method='GET', url=value)
        _uuid = params.get('uuid') or params.get('_id')
        if not _uuid:
            if _app_has_endpoint(flask.current_app, self.endpoint):
                raise ValueError("{} is not a recognized URI for {}".format(value, self.endpoint))
            else:
                raise AppError("The endpoint '{}' is not a valid endpoint for this application".format(self.endpoint))
        return _uuid

class URL(String):
    TYPE_NAME = 'URL'
    def package(self, response):
        if response is None:
            return response
        if not self._valid_url(response):
            message = "The URL supplied by the programmer '{}' is not a valid URL".format(response)
            LOGGER.error(message)
        return self._strip_credentials(response)

    def unpackage(self, name, data, method):
        if self.nullable and data is None:
            return None
        if not self._valid_url(data):
            raise _raise_validation_type_error(name, 'URL', data)
        return data

    @staticmethod
    def _valid_url(url):
        try:
            parsed = urllib.parse.urlparse(url)
            return all((
                bool(parsed.netloc),
                bool(parsed.scheme),
            ))
        except AttributeError:
            return False

    @staticmethod
    def _strip_credentials(url):
        parsed = urllib.parse.urlparse(url)
        netloc = parsed.netloc
        if parsed.username is not None:
            netloc = netloc[len(parsed.username) + 1:]
        if parsed.password is not None:
            netloc = netloc[len(parsed.password) + 1:]
        return urllib.parse.urlunparse((parsed[0], netloc, parsed[2], parsed[3], parsed[4], parsed[5]))

    def get_filter_value(self, name, value):
        if not self._valid_url(value):
            raise ValueError("{} is not a URL".format(value))
        return value



import logging
from flask import current_app
from sepiida.requests import privileged_session
from sepiida.session import current_user_uri


LOGGER = logging.getLogger(__name__)

def search():
    session = privileged_session()
    url = '{}/memberships/?filter[user]={}'.format(current_app.config.get('SEPIIDA_PAO_ROOT'), current_user_uri())
    response = session.request('GET', url)

    if not response.ok:
        LOGGER.error('Error sending request GET %s: %s %s',
            url,
            response.status_code,
            response.text,
        )
        raise Exception('Failed Request, Response: {}, {}'.format(response.status_code, response.text))
    return response.json()

def has_any(group_names):
    memberships = search()['resources']
    session = privileged_session()
    url = '{}/groups/?filter[name]={}'.format(current_app.config.get('SEPIIDA_PAO_ROOT'), ','.join(group_names))
    response = session.request('GET', url)

    if not response.ok:
        LOGGER.error('Error sending request GET %s: %s %s',
            url,
            response.status_code,
            response.text,
        )
        raise Exception('Failed Request, Response: {}, {}'.format(response.status_code, response.text))

    groups = [group['uri'] for group in response.json()['resources']]

    for membership in memberships:
        if membership['group'] in groups:
            return True
    return False



import csv
import io
import json
import urllib.parse

import accept_types

import sepiida.errors


def loads(request):
    content_type = request.headers.get('Content-Type', '')
    if not content_type:
        raise sepiida.errors.InvalidPayload(
            "Your request did not include the 'Content-Type' header and RFC2616 does not specify a default encoding. Since "
            "we cannot know how your request is encoded we do not know how to decode it. Please include an appropriate "
            "Content-Type header"
        )

    try:
        return REQUEST_CONTENT_TYPES[content_type](request)
    except KeyError:
        error = "The content-type header you specified, {}, is not one that we support. Please try one of {}".format(
            content_type,
            ', '.join(REQUEST_CONTENT_TYPES.keys())
        )
        raise sepiida.errors.InvalidPayload(error)
    except ValueError as e:
        raise sepiida.errors.InvalidPayload(
            "Your payload specified a Content-Type of {} but we were unable to decode it as such: {}".format(content_type, e)
        )

def loads_form(request):
    return request.values.to_dict()

def loads_json(request):
    try:
        decoded = request.data.decode('utf-8')
    except ValueError as e:
        raise ValueError("Your request was not UTF-8 encoded: {}".format(e))

    return json.loads(decoded)

def dumps_csv(response, signature):
    signature = signature.signature if signature else None

    buff = io.StringIO()
    writer = csv.writer(buff, dialect='excel')
    if isinstance(response, dict):
        try:
            resources = response['resources']
        except KeyError:
            resources = [response]
    elif isinstance(response, (list, tuple, set)):
        resources = response
    else:
        resources = [response]
    if not resources:
        return ''
    signature = signature or resources[0]
    if signature:
        keys = sorted(signature.keys())
        writer.writerow(keys)
    for row in resources:
        writer.writerow([row[key] for key in keys])
    buff.seek(0)
    return buff.read()

def dumps_form(response, signature): # pylint: disable=unused-argument
    preprocessed = preprocess_form(response)
    return urllib.parse.urlencode(preprocessed)

def dumps_json(response, signature): # pylint: disable=unused-argument
    return json.dumps(response)



REQUEST_CONTENT_TYPES = {
    'application/json'                  : loads_json,
    'application/x-www-form-urlencoded' : loads_form,
}
RESPONSE_CONTENT_TYPES = {
    'application/json'                  : dumps_json,
    'application/x-www-form-urlencoded' : dumps_form,
    'text/csv'                          : dumps_csv,
}
CONTENT_TYPE_PREFERENCE_ORDER = [
    'application/json',
    'application/x-www-form-urlencoded',
    'text/csv',
]
assert set(RESPONSE_CONTENT_TYPES.keys()) == set(CONTENT_TYPE_PREFERENCE_ORDER), (
    "Please ensure that every key in REQUEST_CONTENT_TYPES is also in CONTENT_TYPE_PREFERENCE_ORDER"
)

def prepare_response(request, response, headers, signature):
    content_type = request.headers.get('Content-Type') or 'application/json' # Sometimes we get empty strings for Content-Type
    match = accept_types.get_best_match(request.headers.get('Accept', content_type), CONTENT_TYPE_PREFERENCE_ORDER)
    if not match:
        raise sepiida.errors.UnableToSerialize(title=(
            "No serializers are able to respond in a format you would accept. "
            "Check your Accept header and alter it to accept more types for this request. "
            "Possible MIME types include: {}".format(", ".join(RESPONSE_CONTENT_TYPES.keys()))
        ))
    new_headers = headers.copy()
    new_headers['Content-Type'] = match
    return RESPONSE_CONTENT_TYPES[match](response, signature), new_headers

def preprocess_form(data):
    "This should covert a nested dictionary with lists into a flat dictionary in preparation for urlencoding, recursively"
    results = {}
    if isinstance(data, dict):
        for k, v in data.items():
            if isinstance(v, (int, float, str, type(None))):
                results[k] = v
            else:
                preprocessed = preprocess_form(v)
                for subk, subv in preprocessed.items():
                    results["{}{}".format(k, subk if subk.startswith('[') else "." + subk)] = subv
    else:
        for i, d in enumerate(data):
            if isinstance(d, (int, float, str, type(None))):
                results["[{}]".format(i)] = d
            else:
                for subk, subv in preprocess_form(d).items():
                    results["[{}].{}".format(str(i), subk)] = subv
    return results



import collections
import json
import logging
import uuid

import boto3.session

LOGGER = logging.getLogger(__name__)

class Queue(): # pylint: disable=too-few-public-methods
    def __init__(self, queuename, access_key_id=None, secret_access_key=None, region_name=None):
        region_name = region_name or 'us-west-2'
        self.session = boto3.session.Session(
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            region_name=region_name)

        self.sqs = self.session.resource('sqs')
        self.queue = self.sqs.get_queue_by_name(QueueName=queuename)

    def push(self, credentials, content):
        message = {
            'credentials'   : credentials,
            'content'       : content,
        }
        result = self.queue.send_message(MessageBody=json.dumps(message))
        LOGGER.info("Created message with message ID %s and MD5 %s",
            result.get('MessageId'),
            result.get('MD5OfMessageBody'),
        )
        return result.get('MessageId')

    def messages(self, wait_time_seconds=20):
        for message in self.queue.receive_messages(WaitTimeSeconds=wait_time_seconds):
            yield Message(message)

class Message(): # pylint: disable=too-few-public-methods
    def __init__(self, message):
        self.message = message
        self.body = json.loads(message.body)

    def delete(self):
        return self.message.delete()

    def __getattr__(self, name):
        return getattr(self, name)

class SQSFake(): # pylint: disable=too-few-public-methods
    FakeMessage = collections.namedtuple('FakeMessage', ['body'])
    def __init__(self):
        self.sent_messages = []

    def send_message(self, MessageBody):
        self.sent_messages.append(json.loads(MessageBody))
        return {
            'MessageId'         : uuid.uuid4(),
            'MD5OfMessageBody'  : None,
        }

    def receive_messages(self, WaitTimeSeconds=None): # pylint: disable=unused-argument
        for message in self.sent_messages:
            yield self.FakeMessage(body=json.dumps(message))

class QueueFake(Queue):
    def __init__(self): # pylint: disable=super-init-not-called
        self.queue = SQSFake()

    @property
    def sent_messages(self):
        return self.queue.sent_messages

    def sent(self, message):
        return message in self.sent_messages



import contextlib
import os
import re
import subprocess
import sys

from setuptools import setup
from setuptools.command.sdist import sdist

DATA_ROOTS = []
PROJECT = 'sepiida'
VERSION_FILE = 'sepiida/__init__.py'

def _get_output_or_none(args):
    try:
        return subprocess.check_output(args).decode('utf-8').strip()
    except subprocess.CalledProcessError:
        return None

def _get_git_description():
    return _get_output_or_none(['git', 'describe'])

def _get_git_branches_for_this_commit():
    branches = _get_output_or_none(['git', 'branch', '-r', '--contains', 'HEAD'])
    split = branches.split('\n') if branches else []
    return [branch.strip() for branch in split]

def _is_on_releasable_branch(branches):
    return any([branch == 'origin/master' or branch.startswith('origin/hotfix') for branch in branches])

def _git_to_version(git):
    match = re.match(r'(?P<tag>[\d\.]+)-(?P<offset>[\d]+)-(?P<sha>\w{8})', git)
    if not match:
        version = git
    else:
        version = "{tag}.post0.dev{offset}".format(**match.groupdict())
    return version

def _get_version_from_git():
    git_description = _get_git_description()
    git_branches = _get_git_branches_for_this_commit()
    version = _git_to_version(git_description) if git_description else None
    if git_branches and not _is_on_releasable_branch(git_branches):
        sys.stderr.write((
            "Forcing version to 0.0.1 because this commit is on branches {} "
            "and not a whitelisted branch\n").format(git_branches))
        version = '0.0.1'
    return version

VERSION_REGEX = re.compile(r'__version__ = "(?P<version>[\w\.]+)"')
def _get_version_from_file():
    with open(VERSION_FILE, 'r') as f:
        content = f.read()
    match = VERSION_REGEX.match(content)
    if not match:
        raise Exception("Failed to pull version out of '{}'".format(content))
    version = match.group(1)
    return version

@contextlib.contextmanager
def write_version():
    version = _get_version_from_git()
    if version:
        with open(VERSION_FILE, 'r') as version_file:
            old_contents = version_file.read()
        with open(VERSION_FILE, 'w') as version_file:
            new_contents = '__version__ = "{}"\n'.format(version)
            version_file.write(new_contents)
    print("Wrote {} with {}".format(VERSION_FILE, new_contents))
    yield
    if version:
        with open(VERSION_FILE, 'w') as version_file:
            version_file.write(old_contents)
            print("Reverted {} to old contents".format(VERSION_FILE))

def get_version():
    file_version = _get_version_from_file()
    git_version = _get_version_from_git()
    return (file_version == 'development' and git_version) or file_version

def get_data_files():
    data_files = []
    for data_root in DATA_ROOTS:
        for root, _, files in os.walk(data_root):
            data_files.append((os.path.join(PROJECT, root), [os.path.join(root, f) for f in files]))
    return data_files

class CustomSDistCommand(sdist): # pylint: disable=no-init
    def run(self):
        with write_version():
            sdist.run(self)

def main():
    if sys.argv[1] == 'version':
        print(get_version())
        return 0
    setup(
        name                 = "sepiida",
        version              = get_version(),
        description          = "The root library, bringer of common code and goodness",
        url                  = "https://bitbucket.org/Authentise/sepiida.git",
        long_description     = open("README.md").read(),
        author               = "Authentise, Inc.",
        author_email         = "engineering@authentise.com",
        cmdclass             = {
            'sdist'         : CustomSDistCommand,
        },
        install_requires    = [
            'accept_types==0.3.0',
            'arrow==0.8.0',
            'attrdict==2.0.0',
            'blinker==1.4',
            'boto3>=1.3.0',
            'botocore>=1.4.38',
            'celery==3.1.23',
            'cryptography==1.4',
            'Flask==0.11.1',
            'Flask-Cors==2.0.1',
            'Flask-HTTPAuth==2.5.0',
            'Flask-UUID==0.2',
            'PyJWT==1.4.0',
            'raven==5.8.1',
            'requests==2.7.0',
            'PyYAML==3.11',

                'coverage==3.7.1',
                'HTTPretty>=0.8.10',
                'pytest-cov>=2.3.1',
                'pytest-flask>=0.10.0',
                'pytest-mock>=1.2',
                'pytest>=3.0.2',
        ],
        extras_require={
            'develop': [
                'astroid>=1.4.8',
                'mothermayi>=0.4',
                'mothermayi-pylint>=0.5',
                'mothermayi-isort>=0.8',
                'pylint>=1.5.5',
            ]
        },
        packages             = ["sepiida"],
        package_data         = {
            "sepiida"        : ["sepiida/*"],
        },
        data_files           = get_data_files(),
        entry_points = {
            'pytest11': [
                'sepiida = sepiida.fixtures'
            ]
        },
        scripts = [],
        include_package_data = True,
    )
    return 0

if __name__ == "__main__":
    sys.exit(main())



import functools
import importlib
import logging

import celery
import flask
import kombu
import kombu.exceptions
import raven
import raven.contrib.celery

import sepiida.context

LOGGER = logging.getLogger(__name__)

def create(service_name, imports=None, rabbit_url='amqp://', sentry_dsn=None, testing=False, flask_app=None):
    # This little gem of a hack ensures our tasks are getting imported.
    # Normally celery would do this for us, but I think we broke it
    # because of the way we are wrapping celery tasks
    if imports:
        _ = [importlib.import_module(_import) for _import in imports]

    app = celery.Celery(service_name)
    app.conf.update(
        BROKER_URL = rabbit_url,
        CELERY_DEFAULT_QUEUE = service_name,
        CELERY_QUEUES = (
            kombu.Queue(service_name, kombu.Exchange(service_name), routing_key=service_name),
        ),
        CELERY_TIMEZONE = 'UTC',
    )

    if testing:
        app.conf.update(CELERY_ALWAYS_EAGER=True)

    if sentry_dsn:
        client = raven.Client(sentry_dsn)
        raven.contrib.celery.register_signal(client)

    create.application = app
    create.flask_app = flask_app

    TaskProxy.register_all()

    return app
create.application = None
create.flask_app = None

class TaskProxy(object): #pylint: disable=too-few-public-methods
    instances = []
    def __init__(self, func, *args, schedule=None, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.celery_task = None
        self.schedule = schedule
        TaskProxy.instances.append(self)

    @staticmethod
    def register_all():
        for instance in TaskProxy.instances:
            instance.ensure_registered()

    def ensure_registered(self):
        if not self.celery_task:
            base = self.kwargs.pop('base', None)
            if base:
                raise Exception("You can't change the base class, sorry")
            self.celery_task = create.application.task(*self.args, base=TaskWithCredentials, **self.kwargs)(self.func)
            if self.schedule:
                create.application.conf['CELERYBEAT_SCHEDULE'][self.name] =  {
                    'task'      : self.name,
                    'schedule'  : self.schedule,
                }

    def __call__(self, *args, **kwargs):
        self.ensure_registered()
        return self.celery_task(*args, **kwargs)

    def __getattr__(self, name):
        self.ensure_registered()
        return getattr(self.celery_task, name)

def has_task_context():
    return task_credentials.current_task is not None

def task_credentials():
    if not has_task_context():
        raise Exception("Cannot retrieve task credentials - not in a task context")
    return task_credentials.current_task.credentials
task_credentials.current_task = None

class TaskWithCredentials(celery.Task): # pylint: disable=abstract-method
    @staticmethod
    def _add_credentials_headers(options):
        if has_task_context():
            credentials = task_credentials()
        else:
            credentials = sepiida.context.extract_credentials(flask.request)

        LOGGER.debug("Extracted credentials %s", credentials)
        options['headers'] = options.get('headers', {})
        options['headers'].update({
            'credentials': credentials,
        })

    def apply_async(self, args=None, kwargs=None, task_id=None, producer=None, link=None, link_error=None, **options):
        self._add_credentials_headers(options)
        return super().apply_async(
            args       = args,
            kwargs     = kwargs,
            task_id    = task_id,
            producer   = producer,
            link       = link,
            link_error = link_error,
            **options
        )

    def retry(self, args=None, kwargs=None, exc=None, throw=True, eta=None, countdown=None, max_retries=None, **options): # pylint: disable=too-many-arguments
        self._add_credentials_headers(options)
        return super().retry(
            args        = args,
            kwargs      = kwargs,
            exc         = exc,
            throw       = throw,
            eta         = eta,
            countodwn   = countdown,
            max_retries = max_retries,
            **options
        )

    @property
    def credentials(self):
        return self.get_header('credentials')

    def get_header(self, key, default=None):
        return (self.request.headers or {}).get(key, default)

def task(*args, **kwargs):
    def decorator(func):
        @functools.wraps(func)
        def __add_task_context(*args, **kwargs):
            if len(args) > 0 and isinstance(args[0], TaskWithCredentials):
                _task = args[0]
            else:
                _task = None
            try:
                task_credentials.current_task = _task
                return func(*args, **kwargs)
            finally:
                task_credentials.current_task = None
        return TaskProxy(__add_task_context, *args, **kwargs)
    return decorator

def with_app_context(func):
    @functools.wraps(func)
    def __inner(*args, **kwargs):
        with create.flask_app.app_context():
            return func(*args, **kwargs)
    return __inner



import logging

LOGGER = logging.getLogger(__name__)

def stabilize_werkzeug_iterator():
    """
    Once upon a time we created a bunch of microservices. These services often had to deal with problems.
    So we connected them to sentry to log those problems. That was great until we had to deal with the fact
    that werkzeug actually gets an OSError on the socket when a client does the following:

    1. Connect to the service.
    2. Request a resource
    3. Get disconnected, either by force or by choice, before the server sends all the data for the request

    What is going on is that the OS-level socket is now closed and we can't send data into it. This causes
    an OSError in the python level inside werkzeug. This is an unhandled exception. Because the connection
    is already closed it doesn't matter to werkzeug that this is an unhandled exception - the transaction
    is done no matter what. To us, with our integration with sentry, we *do* care because we treat all
    unhandled exceptions as programmer error. We are finally taking a stand and saying this is not *our*
    error.
    """
    import werkzeug.wsgi
    def stable_next(self):
        try:
            return self._next() # pylint: disable=protected-access
        except OSError as e:
            LOGGER.warning(
                "Caught OSError when iterating over WSGI socket (%s). "
                "This usually means the client disconnected before we were done sending them data.", e
            )
            raise StopIteration(str(e))
    werkzeug.wsgi.ClosingIterator.__next__ = stable_next



import urllib.parse

import flask
import werkzeug.exceptions
import werkzeug.routing

import sepiida.errors


class FakeRequest(): # pylint: disable=too-few-public-methods
    def __init__(self, method, url, app):
        self.environ = {
            'CONTENT_LENGTH'    : '0',
            'CONTENT_TYPE'      : '',
            'HTTP_HOST'         : app.config.get('SERVER_NAME') or 'localhost',
            'PATH_INFO'         : url,
            'QUERY_STRING'      : '',
            'REQUEST_METHOD'    : method,
            'SERVER_NAME'       : app.config.get('SERVER_NAME') or 'localhost',
            'wsgi.url_scheme'   : 'http',
        }

def extract_parameters(app, method, url):
    if url.startswith('http'):
        parsed = urllib.parse.urlparse(url)
        url = parsed.path
    request = FakeRequest(method, url, flask.current_app)
    adapter = app.create_url_adapter(request)
    try:
        return adapter.match()
    except (werkzeug.exceptions.NotFound, werkzeug.exceptions.MethodNotAllowed):
        return None, {}
    except werkzeug.routing.RequestRedirect as e:
        return extract_parameters(app, method, e.new_url)

def uri(endpoint, uuid=None, method='GET', **kwargs):
    try:
        return flask.url_for(endpoint, uuid=uuid, _external=True, _method=method, _scheme='https', _single_resource=True, **kwargs)
    except werkzeug.routing.BuildError:
        error = (
            "Failed to build a URI. You may have attempted to create a URI on a resource ({}) "
            "that does not have a get() method or supplied a UUID when it expects an integer ID or vice-versa."
        ).format(endpoint)
        raise sepiida.errors.RoutingError(error)
